/*
 * @(#) core.h 23.2 90/06/19 SMI
 */

/*
 * Copyright (c) 1988 by Sun Microsystems, Inc.
 */

/*
 * core.h
 *	Definitions for the interface to the core layer of the server.
 *	The core layer manipulates the units of work on which the
 *	server operates: creating, destroying, tracking references to
 *	them, etc.  The information here is public except where noted.
 */

#ifndef _core_h
#define _core_h

#include <palloc.h>
#include "defs.h"
#include "obj_types.h"

/*
 * REF_FIELDS - The fields that comprise a ref.  Used to define the 
 *		ref data structure and within the server object
 *		structure.	
 *
 * 	Note that the bit fields are carefully laid out to take up 
 * 	exactly 32 bits.  Their bit positions within the word are also
 * 	carefully set to avoid a shift when constructing an address
 *	from the entity offset field.
 */
#ifndef LITTLEENDIAN
/* 	Big Endian machines: ex. 68K */

#define REF_FIELDS                                                    \
    enum type   type:6;         /* Type of the referenced object. */  \
    unsigned    entity:24;      /* If set, pointer offset to optional \
                                 * entity. */ 			      \
    unsigned    executable:1;	/* If true, PostScript object being   \
				 * referenced is executable.  Here for\
				 * compactness. */		      \
    unsigned    soft:1;         /* If true, this reference is soft. */

#else	/* LITTLEENDIAN */
/* 
 * 	For Little Endian machines (80386, VAX), the order of the 
 *	bitfields is reversed so shifting and masking still work.  
 */

#define REF_FIELDS                                                    \
    unsigned    soft:1;         /* If true, this reference is soft. */\
    unsigned    executable:1;	/* If true, PostScript object being   \
				 * referenced is executable.  Here for\
				 * compactness. */		      \
    unsigned    entity:24;      /* If set, pointer offset to optional \
                                 * entity. */ 			      \
    enum type   type:6;         /* Type of the referenced object. */

#endif	/* LITTLEENDIAN */

/*
 * The ref data structure.
 * It is the principal lvalue used throughout the server to
 * reference a shared object.
 */
struct ref {
    REF_FIELDS
};

/*
 * Typedef for use in server declarations.
 */
typedef struct ref REF;

/*
 * ENTITY_FIELDS - The fields that comprise an entity.  Used to define
 *		   the entity data structure and within the server data
 *		   structures that represent shared objects.
 *
 *	Note that the first four bit fields are carefully laid out to
 *	take up exactly 32 bits.  Their bit positions within the word
 *	are also carefully set to avoid a shift when constructing an
 *	address from the instancedict offset field.
 */

#ifndef LITTLEENDIAN
/* 	Big Endian machines: ex. 68LK */

#define ENTITY_FIELDS 						  	      \
									      \
    enum type	type:6;			/* Type of associated object.*/       \
    int 	instancedict:24;	/* If set, pointer offset to          \
					 * optional dictionary. */	      \
    unsigned	readable:1;		/* If true, entity is readable. */    \
    unsigned	writeable:1;		/* If true, entity is writeable. */   \
    unsigned	refcount:16;		/* Total # of references to entity. */\
    unsigned	softcount:16;		/* # of soft references to entity. */

#else	/* LITTLEENDIAN */
/* 
 *	For Little Endian machines (80386, VAX), the order of the first
 *	four bitfields is reversed so shifting and masking still work.  
 */

#define ENTITY_FIELDS 						  	      \
    unsigned	writeable:1;		/* If true, entity is writeable. */   \
    unsigned	readable:1;		/* If true, entity is readable. */    \
    int 	instancedict:24;	/* If set, pointer offset to          \
					 * optional dictionary. */	      \
    enum type	type:6;		/* Type of associated object.*/       \
    unsigned	refcount:16;		/* Total # of references to entity. */\
    unsigned	softcount:16;		/* # of soft references to entity. */

#endif	/* LITTLEENDIAN */

/*
 * The entity data structure.
 * Usually used in this form by core processing.  Pointers to
 * this data structure are generally NOT used to reference a
 * shared server object - a REF is used.
 */
struct entity {
    ENTITY_FIELDS
};

/*
 * Typedef for use in server declarations.
 */
typedef struct entity ENTITY;

/*
 * Public function declarations
 */
extern REF	new( /* enum type type */ );
extern REF	pnew( /* Pool *pool, enum type type, Uns32 size */ );
extern void	register_type( /* enum type type, char *name,
			unsigned size, Pool *pool,
			int (*init_proc)(), void (*destroy_proc)(), 
			void (*obsolete_proc)() */ );

#define CREATE_TYPE_POOL	((struct pool_desc *)-1)
#define NO_POOL			((struct pool_desc *)NULL)

/*
 * Mask and constant definitions.
 */

/*
 * Mask used in comparison operations to mask off the sfot and
 * executable bits in a REF.
 */
#define ValidRefBits		(~3)
/*
 * Mask used to turn an offset value into an address: used for the
 * entity offset in a REF and the instancedict offset in an ENTITY.
 *
 * NOTE that the bit fields are laid out in both so the positions
 * of the offset fields match.
 */
#define ADDR_MASK 		0x03FFFFFC
/*
 * The highest possible refcount before overflow of the count.
 */
#define MAXIMUM_REFCOUNT 	((1<<16) - 1)
/*
 * The amount a 32 bit quantity is shifted to access the 6 type bits
 * in either a REF or an ENTITY, and the amount that the high order
 * byte of a REF or an ENTITY is shifted to access the type.
 *
 * NOTE that the bit fields are laid out in both so the positions of
 * the type fields match.
 */
#define TYPE_SHIFT		26
#define BYTE_TYPE_SHIFT		2
/*
 * A global REF that contains all zeroes.  Used for comparisons,
 * return values, initialization, etc.
 */
extern REF	NULLREF;

/*
 * Utility macro definitions.
 * Block comments are used here to aid in reading the subsequent
 * code.  Care must be exercised when editing this file to keep
 * the comments current.
 */

/*
 * refword ( REF *refp )
 *	Given the address of a REF, convert it to a 32-bit unsigned
 *	value to access the 32 bits.
 * typebyte ( REF *refp )
 *	Given the address of a REF, fetch the unsigned byte containing
 *	the type bits.  Type comparisons and tests can be performed on
 *	this entire byte for the non-entitied types since the offset
 *	field of the REF will be zero for those types.
 * zero_ref ( REF ref )
 *	Set the REF to null.
 * is_ref_zero ( REF ref )
 *	Return a boolean, indicating whether the REF is null.
 * is_ref_type ( REF ref, enum type type )
 *	Return a boolean, indicating whether the REF is null.
 * ref_is_soft ( REF ref )
 *	Return a boolean, indicating whether the reference represented
 *	by the REF is soft.
 * equal_ref ( REF ref1, REF ref2 )
 *	Return a boolean, indicating whether the type and entity
 *	offset fields match.  NOTE that the soft and executable
 *	bits are ignored.
 * set_entity_of ( REF ref, ENTITY *entp )
 *	Given a REF and the address of an ENTITY, set the entity 
 *	offset field in the REF to reference the ENTITY.  NOTE that
 *	the type, soft and executable bits are not affected by this
 *	operation.
 * assign_ref ( REF ref, ENTITY *entp )
 *	Given a REF and the address of an ENTITY, set the type and
 *	entity offset fields in the REF to refer to the ENTITY.
 *	NOTE that the soft and executable bits are zeroed.
 * entity_addr ( REF ref )
 *	Given a REF, return the ADDRESS of the ENTITY to which the 
 *	REF refers.
 * entity_of ( REF ref )
 *	Given a REF, return the ENTITY to which the REF refers.  NOTE 
 *	that this does NOT return a pointer.
 * refcount_of ( REF ref )
 *	Given a REF, return the total number of references to the 
 *	ENTITY referenced by the REF.
 * softcount_of ( REF ref )
 *	Given a REF, return the number of soft references to the 
 *	ENTITY referenced by the REF.
 * ref_has_entity ( REF ref )
 * 	Return a boolean, indicating whether the REF refereneces an
 *	object that has an ENTITY at its root.
 * entity_has_refcount ( REF ref )
 *	Return a boolean, indicating whether the REF references an
 *	object that is refcounted.
 */
#define refword(refp) 		(*((long *)(refp)))
#ifndef LITTLEENDIAN
#define typebyte(refp) 		(*((unsigned char *)(refp)))
#else
#define typebyte(refp) 		(*(((unsigned char *)(refp))+3))
#endif
#define zero_ref(ref) 		(refword(&(ref)) = 0)
#define zero_object(objp)	(((long *)(objp))[0] = 0, ((long *)(objp))[1] = 0)
#define is_ref_zero(ref) 	(refword(&(ref)) == 0)
#define is_ref_type(ref, type)					\
	(typebyte(&(ref)) == ((unsigned char)type)<<BYTE_TYPE_SHIFT)
#define ref_is_soft(ref) 	((ref).soft)
#define equal_ref(ref1, ref2)  					\
	(((refword(&(ref1))^refword(&(ref2)))&ValidRefBits) == 0)
#ifdef DEBUG
#define set_entity_of(ref, entp)				\
	(assert(((long)entp & (long)~ADDR_MASK) == 0),		\
	 (refword(&(ref)) = refword(&(ref)) & 			\
			~ADDR_MASK | (long)(entp)))
#else	/* ! DEBUG */
#define set_entity_of(ref, entp)				\
	(refword(&(ref)) = refword(&(ref)) & 			\
			~ADDR_MASK | (long)(entp))
#endif	/* DEBUG */
#define assign_ref(ref, entp) 					\
	((refword(&(ref)) = ((unsigned)(entp)->type) << 	\
					TYPE_SHIFT),		\
	 (set_entity_of((ref), (entp))))
#define entity_addr(ref) 					\
	((ENTITY *)(refword(&(ref)) & ADDR_MASK))
#define entity_of(ref) 		(*(entity_addr(ref)))
#define refcount_of(ref)	(entity_of(ref).refcount)
#define softcount_of(ref)	(entity_of(ref).softcount)
#define ref_has_entity(ref)					\
	((unsigned)((ref).type) >= (unsigned)ENTITY_TYPE)
#define entity_has_refcount(ref)				\
	(((unsigned)(ref).type) >= (unsigned)REFCOUNTED_TYPE)
#ifdef USE_NEW_TYPE_STUFF
#define ref_has_entity(ref)					\
	(typebyte(&(ref)) >= ((unsigned char)ENTITY_TYPE)<<BYTE_TYPE_SHIFT)
#define entity_has_refcount(ref)				\
	(typebyte(&(ref)) >= ((unsigned char)REFCOUNTED_TYPE)<<BYTE_TYPE_SHIFT)
#endif

/*
 * Private function declarations
 */
extern int call_destroy_proc( /* REF ref */ );
extern int call_obsolete_proc( /* REF ref */ );

#ifdef REF_TRACE
#define IF_REF_TRACE(x) x
#else
#define IF_REF_TRACE(x)
#endif

/*
 * If REFCOUNT_BY_FUNCTION is defined, the macros incref, decref, 
 * harden and soften become functions.  Used for debugging and
 * gathering statistics.  The macros are named in caps and the
 * functions in lowercase so that both can be used in one server
 * to maximize performance.
 * NOTE that the functions in core.c must match the macros EXACTLY.
 */

/*
 * incref ( REF ref )
 *	Increment an entity's total reference count.  If the
 *	reference is soft, the number of soft references is
 *	also incremented.
 *
 * LIMITATION: If the total refcount of the entity is equal to
 *	MAXIMUM_REFCOUNT, the total refcount of the entity will
 *	NOT be incremented.
 */
#define INCREF(ref) {							\
    if (entity_has_refcount(ref)) { 					\
	register ENTITY *entp = entity_addr(ref);			\
	assert(entp->softcount <= entp->refcount); 			\
	IF_REF_TRACE(trace_incref(entp));				\
	if (entp->refcount != MAXIMUM_REFCOUNT) {			\
	    entp->refcount++;						\
	    if (ref_is_soft(ref)) entp->softcount++; 			\
	}								\
    } 									\
}

/*
 * decref ( REF ref )
 *	Decrement an entity's total reference count.
 *	If the resultant toal count is 0, the associated object is
 *	    destroyed (details of destruction processing depend on
 *	    object's type).
 *	If the reference is soft, the entity's count of soft references
 *	    is also decremented.
 *	If the resultant total count equals the soft count, obsolete
 *	    notification is sent.
 *
 * NOTE that the second assert does the following:
 *	If the reference is soft, both the soft and total counts are
 *	decremented; so initially the softcount must be <= total count.
 *	If the reference is hard, only the toal count is decremented;
 *	so initially the softcount must be < total count.
 * LIMITATION: Any object whose total count is more than MAXIMUM_REFCOUNT 
 *	will not be processed (and will therefore not be destroyed).
 */
#define DECREF(ref) {							\
    if (entity_has_refcount(ref)) { 					\
	register ENTITY *entp = entity_addr(ref);			\
	assert(entp->refcount != 0); 					\
	assert(ref_is_soft(ref) ? (entp->softcount <= entp->refcount)	\
		: (entp->softcount < entp->refcount));			\
	IF_REF_TRACE(trace_decref(entp));				\
	if (entp->refcount != MAXIMUM_REFCOUNT) { 			\
	    if ((--entp->refcount) == 0)				\
		call_destroy_proc(ref); 				\
	    else if (ref_is_soft(ref))					\
		entp->softcount--;					\
	    else if (entp->softcount == entp->refcount)			\
		call_obsolete_proc(ref); 				\
	} 								\
    }									\
}

/*
 * harden ( REF *refp )
 *	Make a soft reference hard.
 *	An address is used because this processing alters a REF field.
 *	The soft count of the associated entity is decremented
 *	    because a soft reference has been removed.
 */
#define HARDEN(refp) {							\
    if (entity_has_refcount(*(refp)) && (refp)->soft) {			\
	register ENTITY *entp = entity_addr(*(refp));			\
	assert(entp->refcount > 0); 					\
	assert(entp->softcount <= entp->refcount); 			\
	IF_REF_TRACE(trace_harden(entp));				\
	if (entp->softcount != MAXIMUM_REFCOUNT) { 			\
	    entp->softcount--; 						\
	    (refp)->soft = 0; 						\
	} 								\
    } 									\
}

/*
 * soften ( REF *refp )
 *	Make a hard reference soft.
 *	An address is used because this processing alters a REF field.
 *	The soft count of the associated entity is incremented
 *	    because a soft reference has been added.
 *	If the resultant soft count equals the total refcount,
 *	    obsolete notification is sent.
 */
#define SOFTEN(refp) { 							\
    if (entity_has_refcount(*(refp)) && !(refp)->soft) {		\
	register ENTITY *entp = entity_addr(*(refp));			\
	assert(entp->refcount > 0); 					\
	IF_REF_TRACE(trace_soften(entp));				\
	if (entp->softcount != MAXIMUM_REFCOUNT) { 			\
	    entp->softcount++; 						\
	    (refp)->soft = 1;						\
	    if (entp->softcount == entp->refcount) 			\
		call_obsolete_proc(*(refp)); 				\
	} 								\
	assert(entp->softcount <= entp->refcount); 			\
    }									\
}

#ifndef REFCOUNT_BY_FUNCTION

#define incref INCREF
#define decref DECREF
#define harden HARDEN
#define soften SOFTEN

#endif	/* REFCOUNT_BY_FUNCTION */

/* The following structure shouldn't be here: it's just here to break
 * a circular reference.  It should really be in cs_common.h */
struct packedcolor {		/* A description of a color */
    unsigned char red, green, blue;
};

/*
 * struct object - A non-refcounted object - no associated corpus struct or
 *		   representation at other system levels; no entity.  This is 
 *		   the structure used to represent items on the operand stack.
 *		   REF_FIELDS are defined at the core layer: type and executable
 *		   are the only ones used for non-refcounted object types.
 *	NOTE: marker and null types have no corresponding value fields.
 *	NOTE: string, array and file fields are only set when the object is
 *	associated with an entitied object of that type.
 *		
 */
struct object {
    REF_FIELDS				/* reference information */
    union {
	long	   integer;		/* type == integer_type */
	long       fixed;		/* type == fixed_type, boolean_type */
		/* the "fract" type should really have been used here
		 * instead of "long", but it's defined in a header that won't
		 * be loaded for a while (don't you hate circular references?) */
	float       real;		/*      == real_type */
	struct operator_def *def;	/*      == operator_type, 
					 * 	   called_operator_type, 
					 *	   magic_variable_type */
	struct {			/*      == string_type */
	    unsigned short start,
	                   length;
	} substring;
	struct {			/*      == array_type */
	    unsigned short start,
	                   length;
	} subarray;
	struct {			/*	== save_type */
	    int level;
	} save;

	struct packedcolor color;	/* type == color_type */
    } value;
};

#endif	/* #ifndef _core_h */
