/*
 * @(#)cs_font.h 23.3 90/06/19
 */

/*
 * Copyright (c) 1984 by Sun Microsystems, Inc.
 */

/*-
	Font related declarations

 */

/*-	The following diagram explains some of the important vectors in
	the definition of a glyph
   (0,0)
   *-----------------+
   |                 |
   |                 |
   |                 |
   |                 |
   |                 |
   |                 |
   |                 |
   |  origin         |
   |  *--------------+--->*width[relative to the origin]
   |                 |
   |                 |
   |                 |
   +-----------------*size


	A font is written to a file as a font struct followed by all the
	glyphs.  All the pointers are relative to the beginning of the
	font.  When the font is read in or written out the pointers are
	converted from and to this offset form.

	The wglyph structure is separate from the glyph structure because
	we need to be able to build remapped versions of a font that differ
	only in widths.
 */


#ifndef _cs_font_h
#define _cs_font_h

#include <graphics/shapes.h>	/* Necessary to get definition of */
				/* fract, spoint, and fpoint. */
#include <Core/core.h>

enum FontType {
    LRbitmapFont,		/* Bitmap with rows left-to-right */
    WidthFont,			/* Font with only printer width information */
    PrinterWidthFont,		/* Scalable width information that applies to
				 * Adobe PostScript printers */
    SplineFont,			/* Spline outline font */
    VectorFont,			/* Simple vector font */
};

enum FontEncoding {
    ASCIIencoding,		/* 8 bit western European ASCII */
    AdobeEncoding,		/* The Adobe version of ASCII */
    JISC6226Encoding,		/* Japanese Industrial Standard Kanjii */
    CyrillicEncoding,
    HangulEncoding,
    DevenageriEncoding,
    AdobeSymbolEncoding,	/* An Adobe Symbol font */
    DingbatsEncoding,		/* ZapfDingbats */
    UnknownEncoding,		/* Can't figure it out */
    MatrixEncoding		/* min/max byte fields are valid. */
};

struct wglyph {			/* A glyph with its associated width
				 * and origin information */
    GLYPH glyph;		/* The Shapes glyph. */
    struct spoint origin;	/* The origin position vector */
    struct fpoint width;	/* The origin-to-origin width.  Note that this
				 * is in subpixel coordinates. */
};


/*
 * Rendering info describes where a collection of glyphs should be
 * rendered.  Rendering info is constructed by a font's
 * calc_rendering_info method.  The contents are passed on
 * to the Shapes glyph renderer.
 */

struct rendering_info {
    GLYPH_REF * ri_glyph_refs;	/* An array of GLYPH_REFS. */
    Sgn32 ri_nglyphs;		/* How many glyphs to render? */
    POINT_B2D ri_origin;	/* The origin and dimensions describe */
				/* the bounding box of bits that will */
				/* be touched by the render operation. */
    POINT_B2D ri_dimensions;
    POINT_B2D ri_finalpos;		/* The final position. */
    GLYPH_opt ri_optimizations;	/* What Shapes optimizations are */
				/* applicable?  */
};

typedef struct rendering_info RENDERING_INFO;

#define ri_get_glyph_refs(xxx) ((xxx)->ri_glyph_refs)
#define ri_set_glyph_refs(xxx,yyy) ((xxx)->ri_glyph_refs = (yyy))

#define ri_get_nglyphs(xxx) ((xxx)->ri_nglyphs)
#define ri_set_nglyphs(xxx,yyy) ((xxx)->ri_nglyphs = (yyy))

#define ri_get_optimizations(xxx) ((xxx)->ri_optimizations)
#define ri_set_optimizations(xxx,yyy) ((xxx)->ri_optimizations = (yyy))


/*
 * Handle these in a special way: return pointers to the points
 * instead of the points themselves.
 */

#define ri_get_origin(RI)		(&((RI)->ri_origin))
#define ri_get_dimensions(RI)		(&((RI)->ri_dimensions))
#define ri_get_finalpos(RI)		(&((RI)->ri_finalpos))

#define DrawsLeftToRight	0	/* Values for drawDirection */
#define DrawsRightToLeft 	1

typedef struct {
    Unsgn8 	drawDirection;
    Sgn16	fontAscent;
    Sgn16	fontDescent;
    Sgn16	overallAscent;
    Sgn16	overallDescent;
    Sgn32	overallWidth;
    Sgn32	overallLeft;
    Sgn32	overallRight;
} TEXTEXTENT_INFO;

typedef struct {
    Sgn16 leftSideBearing;
    Sgn16 rightSideBearing;
    Sgn16 characterWidth;
    Sgn16 ascent;
    Sgn16 descent;
    Unsgn16 attributes;
} XCHAR_METRIC;

typedef struct {
    Unsgn32 name;
    Unsgn32 value;
} XFONT_PROP;

typedef struct {
    XCHAR_METRIC minBounds;
    XCHAR_METRIC maxBounds;
    Unsgn16 minCharOrByte2, 
	    maxCharOrByte2;
    Unsgn8  minByte1, 
	    maxByte1;
    Unsgn16 defaultChar;
    Unsgn16 nFontProps;
    Unsgn8  drawDirection;
    Unsgn8  allCharsExist;
    Sgn16   fontAscent, 
	    fontDescent;
    Unsgn32 nCharsInfo;
} XFONT_INFO;

struct font_opsvector {		/* Operations on a font */
    Sgn32 (*calc_rendering_info) ();
    void (*calc_stringwidth) ();
    void (*calc_bbox) ();
    Sgn32 (*get_glyphs) ();
};

struct font {
    ENTITY_FIELDS		/* header and ref information */
    union {
	Sgn32         magic;	/* Magic number that identifies this font
				 * format */
	struct font_opsvector* ops;	/* The operation definitions for this
					 * object */
    }           u;
    enum FontType ftype:8;	/* The type of representation used for this
				 * font */
    enum FontEncoding encoding:8;  /* The encoding used for this font
				    * (ASCII, AdobeAscii, Kanjii, Hangul,
				    * Devenageri, Cyrillic...) */
    unsigned int left_to_right:1; /* true iff for all glyphs, wy==0 && wx>=0 */
    unsigned int narrow:1;	/* true iff for all glyphs, sx<=16 */
    unsigned int fixed_width:1;	/* true iff all glyphs have the same width */
    unsigned int printermatched:1; /* true iff the widths in this font
				    * have been matched to the widths on
				    * a printer */
    unsigned int monotonic:1;	/* true iff the signs of the x widths match
				 * for all glyphs and the signs of the y
				 * widths match for all glyphs */
    unsigned int sharedglyphs:1; /* true iff glyphs are shared w/ another font*/
    unsigned int cached:1;	/* true iff font is in the unused cache. */
    unsigned int pad1:9;	/* Pad bitfields to a full word. */
    struct spoint origin,
                  size;		/* The bounding box for all glyphs in the
				 * font.  Think of superimposing all the
				 * glyphs so that their origins coincide. */
    struct fpoint ascent;	/* Logical font ascent (baseline up) */
    struct fpoint descent;	/* Logical font descent (baseline down) */
    XFONT_PROP *properties;	/* X font properties */    
    union {
        Sgn8   *name;		/* The name of this font's family */
  	struct fontfamily *family;/* The family structure for this font */
    } fam;
    union {
        Sgn8   *owner;		/* Ownership information for the font */
	REF	srcfont;	/* The font used to generate this one. */
    } comment;
    Unsgn8 firstCol, lastCol, firstRow, lastRow;
    fract       matrix[3][2];	/* A 3x2 matrix describing this font's 
				 * scaling and rotation */
    Unsgn32	fsize;		/* The size of this font in bytes */
    Unsgn16 	nglyphs;	/* The number of glyphs in this font */
    Unsgn16 	minglyph;	/* the lowest numbered valid glyph.
				 * The glyphs array starts at this
				 * index */
    Unsgn16 	defaultchar;	/* The "default" character -- replaces
				 * all missing characters */
    Sgn16       pad2;		/* Pad out for compatibility with
				 * machines that do long word int alignment */
    struct wglyph glyphs[1];	/* The array of wglyphs that make up this font */
};

typedef struct font* FONT;

#define font_get_type(xxx) ((xxx)->type)
#define font_set_type(xxx,yyy) ((xxx)->type = (yyy))

#define font_get_magic(xxx) ((xxx)->u.magic)
#define font_set_magic(xxx,yyy) ((xxx)->u.magic = (yyy))

#define font_get_encoding(xxx) ((xxx)->encoding)
#define font_set_encoding(xxx,yyy) ((xxx)->encoding = (yyy))

#define font_get_left_to_right(xxx) ((xxx)->left_to_right)
#define font_set_left_to_right(xxx,yyy) ((xxx)->left_to_right = (yyy))

#define font_get_narrow(xxx) ((xxx)->narrow)
#define font_set_narrow(xxx,yyy) ((xxx)->narrow = (yyy))

#define font_get_fixed_width(xxx) ((xxx)->fixed_width)
#define font_set_fixed_width(xxx,yyy) ((xxx)->fixed_width = (yyy))

#define font_get_printermatched(xxx) ((xxx)->printermatched)
#define font_set_printermatched(xxx,yyy) ((xxx)->printermatched = (yyy))

#define font_get_monotonic(xxx) ((xxx)->monotonic)
#define font_set_monotonic(xxx,yyy) ((xxx)->monotonic = (yyy))

#define font_get_origin(xxx) ((xxx)->origin)
#define font_set_origin(xxx,yyy) ((xxx)->origin = (yyy))

#define font_get_size(xxx) ((xxx)->size)
#define font_set_size(xxx,yyy) ((xxx)->size = (yyy))

#define font_get_name(xxx) ((xxx)->name)
#define font_set_name(xxx,yyy) ((xxx)->name = (yyy))

#define font_get_comment(xxx) ((xxx)->comment)
#define font_set_comment(xxx,yyy) ((xxx)->comment = (yyy))

#define font_get_nglyphs(xxx) ((xxx)->nglyphs)
#define font_set_nglyphs(xxx,yyy) ((xxx)->nglyphs = (yyy))

#define font_get_pad(xxx) ((xxx)->pad)
#define font_set_pad(xxx,yyy) ((xxx)->pad = (yyy))

#define font_get_ascent(xxx) ((xxx)->ascent)
#define font_set_ascent(xxx,yyy) ((xxx)->ascent = (yyy))

#define font_get_descent(xxx) ((xxx)->descent)
#define font_set_descent(xxx,yyy) ((xxx)->descent = (yyy))

#define font_get_defaultchar(xxx) ((xxx)->defaultchar)
#define font_set_defaultchar(xxx,yyy) ((xxx)->defaultchar = (yyy))


#define font_get_wglyphs(f) ((f)->glyphs)
#define font_get_the_wglyph(f,i) (&((f)->glyphs[i]))
#define font_get_Xascent(f) (cfloorfr((f)->ascent.y))
#define font_get_Xdescent(f) (-cfloorfr((f)->descent.y))

#define font_calc_rendering_info(f,s,l,cp,m,r,ri) \
    (*(f)->u.ops->calc_rendering_info)(f,s,l,cp,m,r,ri)

#define font_calc_stringwidth(f,str,size,remap,x,y) \
    (*(f)->u.ops->calc_stringwidth)(f,str,size,remap,x,y)

#define font_calc_bbox(f,str,size,remap,org,dim) \
    (*(f)->u.ops->calc_bbox)(f,str,size,remap,org,dim)

/*
 * font magic number 
 */
#define FONT_MAGIC 	0x137A2B45

/*
 * Compress the table of unused fonts.  Invoked because something
 * removed one or more entries from the table.
 */
#define SHRINKCACHETABLE(to, from) {					\
    memcpy((char *)&cachetable[(to)], (char *)&cachetable[(from)],	\
			(cachetablesize - (from)) * sizeof(FONT));	\
    cachemark -= ((from) - (to));					\
}
   
/*
 * Re-calculate the total memory size of the font because a glyph was
 * added.  If the font is cached, un-cache it, because it's in use 
 * again, and its size is changing.
 */
#define FONTSIZE(fp, addsize) {						\
    if (fp->cached) {							\
	register k;							\
									\
	assert(fp->refcount > 0);					\
	fp->cached = 0;							\
	MA_destroy(misc_family, cachedfont, fp->fsize);                 \
	for (k = 0; (k < cachemark) && (cachetable[k] != fp); k++);	\
	assert(k < cachemark);						\
	cachesize -= fp->fsize;						\
	SHRINKCACHETABLE(k, k+1)					\
        MA_create(core_family, cs_font_type, fp->fsize + addsize);      \
    } else {								\
	MA_adjust(core_family, cs_font_type, addsize);			\
    }									\
    fp->fsize += (addsize);						\
}


#define X11_n1dChars(f, remap) ((remap) ? 256 : font_get_nglyphs(f))

#define X11_n2dChars(f)	(((f)->lastCol - (f)->firstCol + 1) * 		\
			 ((f)->lastRow - (f)->firstRow + 1))


#define X11_nChars(psfont, f)						\
    (((f)->encoding == MatrixEncoding) ?				\
     	X11_n2dChars(f) : X11_n1dChars((f), (psfont)->remap))
     
     
/*
 * The collected modifiers that may be applied to a piece of text when it is
 * shown
 */

struct text_modifiers {
    struct fpoint cdelta;	/* Delta applied to the distinguished
				 * character */
    Sgn8        c;		/* The distinguished character */
    struct fpoint alldelta;	/* Delta applied to all characters */
};

extern Pool	*cs_font_pool;
#endif /* _cs_font_h */
