/*
 * @(#)dictionary.h 23.5 90/06/19
 */

/*
 * dictionary code and constants
 */

#ifndef _DICTIONARY_INCLUDE
#define _DICTIONARY_INCLUDE

/*
 * These are the types of dictionaries that we support.
 */
enum dicttype {
    nulldict,			/* null type (not a dict type) */
    refdict,			/* all keys are REF's */
    genericdict			/* all keys are objects */
};

/*
 * The number of keys that an architecture can search for in the time that
 * it takes to hash into a dictionary (big on SPARC because of the cost
 * of doing a mod).
 *
 * Note that this define also will generate a space savings (the bigger, the
 * more space savings).  See define_object_in_dictionary().
 */
#ifdef sparc
#define LINEAR_SIZE_CUTOFF 32		/* sparc */
#else
#ifdef mc68000
#define LINEAR_SIZE_CUTOFF 8		/* 68K */
#else
#define LINEAR_SIZE_CUTOFF 6		/* i386 */
#endif
#endif



/*
 * Dictionary defs
 */

#define CONVERT_TO_GROWABLE_DICT 20     /* Number of key/value pairs */
/* REMIND Some experimentation shows that if the sys dict at this size
 * it 1) will not grow, 2) has a reduced collision rate, 3) has a
 * reasonable empty/full slot ratio. Use dict_stats for tuning.
 */
#define INITIAL_SYSTEMDICT_SIZE 1209
#define INITIAL_GROWABLE_SIZE 6		/* for other dictionaries */
#define DICT_MAX_USE_PERCENTAGE 80	/* At this % full,  grow the dict */
#define INFINITE_DICT_SIZE      5000	/* Well, practically infinite anyway */

/*
 * the amount to trim dictionaries.  TRIM_SLOP is the percentage slop
 * to pad the dict, and TRIM_SLOP_CONST is the minimum addition (always
 * added.
 */
#define TRIM_SLOP		10
#define TRIM_SLOP_CONST 3




/*
 * These are the different types of dictionary key/value pairs.  They are
 * used in a union in DICT (below).
 */
struct refdict {
    REF key;
    struct object value;
};

struct genericdict {
    struct object key;
    struct object value;
};

/*
 * Dictionary quickhashing macros. Dict key lookups are implemented
 * by a 2 level hash: a cheap primary hash which determines if
 * a key is not in a dict and an expensive secondary (the old hash)
 * which determines for sure if the key is present/absent.
 * The primary hash uses a bit table, the idea was taken from csh.
 */
#define	QDHSHSIZ	128		/* in bits, power of 2 */
#define QDHSHMASK	(QDHSHSIZ - 1)
#define QDHSHMUL	243
#define qdhash(a)	(((a >> 2) * QDHSHMUL) & QDHSHMASK)
#define isqdbit(d, hv)	(d[(hv) >> 3] & 1 << ((hv) & 7))	/* bit test */
#define setqdbit(d, hv)	(d[(hv) >> 3] |= 1 << ((hv) & 7))	/* bit set */

/*
 * The dictionary structure. 
 */
struct dictionary {
    ENTITY_FIELDS			/* header and ref info */
    char qdhashtable[QDHSHSIZ/8];	/* Speed up stack lookups */
    unsigned short size;                /* Actual kv slots allocated */
    unsigned short used;                /* Actual slots used */
    unsigned short max_size;            /* Maximum dict size in kv pairs */
    enum dicttype dicttype:8;		/* the type of dictionary */
    unsigned linear:1;			/* hash or linear array? */
    unsigned pad:7;
    union {				/* The dictionary kv pairs */
	struct refdict *refdict;
	struct genericdict *genericdict;
    } objects;
    IntBackPtr  *interest_backptr;      /* the interests that contain
					 * this dictionary */
};

typedef struct dictionary DICT;

#define dict_of(ref) ((DICT *)entity_addr(ref))

/*
 * return the size of the kv pair for this dicttype
 */
#define dictkvsize(type) 						\
	((type == refdict) ? sizeof(struct refdict)			\
			   : sizeof(struct genericdict))

/*
 * these are the types that can fit in a refdict.
 */
extern unsigned char refdict_types[(unsigned) maximum_type];

/*
 * return the optimum dicttype that will contain this object
 */
#define dictkvtype(obj)							\
    (refdict_types[(unsigned)(obj).type] ? refdict : genericdict)


extern struct object *find_object_in_dictionary();
extern struct object *genericdict_find_object_in_dictionary();
extern enum error_type define_object_in_dictionary();
extern void delete_object_from_dictionary();

#ifndef NODICTRELOC
#define MAX_RELOCATIONS_PER_TRIP	20
extern int count_accesses, relocations_this_trip, max_relocations_per_trip;
extern Pool *compound_hot_pool, *compound_cold_pool;
/* REMIND: no return code to indicate failure to allocate memory */
extern void relocate_dict();
#endif /* NODICTRELOC */

/*
 * If heat_dict returns true, the dict->objects should be relocated.
 * Note the test against compound_cold_pool;  we shouldn't test
 * != compound_hot_pool, particularly in case of ALLOCATE_LOCAL.
 */
#ifndef NODICTRELOC
#define heat_dict(dict)						\
	    count_accesses					\
	&&  relocations_this_trip < max_relocations_per_trip	\
	&&  pool_of(dict->objects.refdict)==compound_cold_pool
#else /* NODICTRELOC */
#define heat_dict(dict) 0
#endif /* NODICTRELOC */

/*
 * check_dict_heat can be defined to nothing for cases where
 * FIND_OBJECT_IN_DICTIONARY may be followed closely by growth.
 */
#ifndef NODICTRELOC
#define check_dict_heat(dict)					\
	if (heat_dict(dict)) relocate_dict(dict)
#else /* NODICTRELOC */
#define check_dict_heat(dict)
#endif /* NODICTRELOC */

#ifndef lint

#ifdef PROCEDURIZE_HASH
#define hash_object(hash, obj)  ((hash) = hash_object_proc(obj))
#else   /* PROCEDURIZE_HASH */
#define hash_object(hash,obj)                                   \
        ((hash) = ((object_word0(obj) & ValidObjectBits)        \
                                          ^ object_word1(obj)))
#define hash_index(hash,mod)                                    \
        ((hash) % (mod))
#endif  /* PROCEDURIZE_HASH */

#else   /* lint */

#define hash_object(hash, obj)  ((hash) = hash_object_proc(obj))

#endif  /* lint */



/*
 * find an object (key) in dictionary (dict).  Put result in result.
 * 	struct object key;
 * 	DICT dict;
 *	struct object *result;
 */
#define FIND_OBJECT_IN_DICTIONARY(k, dict, result) {			\
    check_dict_heat(dict);						\
    switch (dict->dicttype) {						\
    case genericdict:							\
	result = genericdict_find_object_in_dictionary(k, dict);	\
	break;								\
    case refdict:							\
	result = 0;							\
	if (dict->linear) {						\
	    register struct refdict *res, *resbegin;			\
									\
	    assert(dict->used <= LINEAR_SIZE_CUTOFF);			\
	    resbegin = dict->objects.refdict;				\
	    res = resbegin + dict->used - 1;				\
	    while (res >= resbegin) {					\
		if (equal_ref(res->key, k)) {				\
		    result = &(res->value);				\
		    break;						\
		}							\
		res--;							\
	    }								\
	} else {							\
	    register    hash;						\
	    register struct refdict *res, *resbegin;			\
	    DECLARE_COLLIDE						\
	    hash_object(hash, &k);					\
	    resbegin = dict->objects.refdict;				\
	    res = &resbegin[hash_index(hash, dict->size)];		\
	    while (!is_ref_zero(res->key)) {				\
		if (equal_ref(res->key, k)) {				\
		    DICT_STAT_REPORT(ds_collide_type, dict,		\
				     INCR_COLLIDE());			\
		    result = &(res->value);				\
		    break;						\
		}							\
		INCR_COLLIDE();						\
		res--;							\
		if (res < resbegin)					\
		    res = &resbegin[dict->size - 1];			\
	    }								\
	}								\
	break;								\
    default:								\
	assert(/* BAD DICTIONARY TYPE */ 0);				\
	result = 0;							\
	break;								\
    }									\
}

/*
 * find a hashed object (key) in dictionary (dict).  Put result in result.
 * 	struct object key;
 *	int hash;
 * 	DICT dict;
 *	struct object *result;
 */
#define FIND_HASHED_OBJECT_IN_DICTIONARY(k, hash, dict, result) {	\
    check_dict_heat(dict);						\
    switch (dict->dicttype) {						\
    case genericdict:							\
	result = genericdict_find_object_in_dictionary(k, dict);	\
	break;								\
    case refdict:							\
	result = 0;							\
	if (dict->linear) {						\
	    register struct refdict *res, *resbegin;			\
									\
	    assert(dict->used <= LINEAR_SIZE_CUTOFF);			\
	    resbegin = dict->objects.refdict;				\
	    res = resbegin + dict->used - 1;				\
	    while (res >= resbegin) {					\
		if (equal_ref(res->key, k)) {				\
		    result = &(res->value);				\
		    break;						\
		}							\
		res--;							\
	    }								\
	} else {							\
	    register struct refdict *res, *resbegin;			\
	    DECLARE_COLLIDE						\
	    resbegin = dict->objects.refdict;				\
	    res = &resbegin[hash_index(hash, dict->size)];		\
	    while (!is_ref_zero(res->key)) {				\
		if (equal_ref(res->key, k)) {				\
		    DICT_STAT_REPORT(ds_collide_type, dict,		\
				     INCR_COLLIDE());			\
		    result = &(res->value);				\
		    break;						\
		}							\
		INCR_COLLIDE();						\
		res--;							\
		if (res < resbegin)					\
		    res = &resbegin[dict->size - 1];			\
	    }								\
	}								\
	break;								\
    default:								\
	assert(/* BAD DICTIONARY TYPE */ 0);				\
	result = 0;							\
	break;								\
    }									\
}

#endif
